(** * Completeness of [LJT] w.r.t. Kripke Semantics *)

Set Implicit Arguments.

Require Import environments.
Require Import extralibrary.
Require Import sublist.
Require Import variable_sets.
Require Import substitution.
Require Import renaming.
Require Import soundness.
Require Import language_syntax.


(** The completeness of [LJT] w.r.t. Kripke semantics follows
   from the Universal Completeness.

   The universal Kripke model consists of contexts as worlds,
   the sub-context relation, and the provability for atomic
   sentences, and term as the constant domain. *)


(** ** The Universal Model *)


(** Monotonicity of forcing relation *)

Lemma weakening_gen' : forall Ga De A eta,
    sub_ctx Ga De ->
    Ga |- A ->
    rename_c eta De |- rename eta A
                                
with weakening_stoup_gen' : forall Ga De A C eta,
    sub_ctx Ga De ->
    Ga ;; A |- C ->
    rename_c eta De ;; rename eta A |- rename eta C.
Proof.
 - intros; apply weakening_gen with Ga; auto. 
   + apply indeed_sub_ctx_is; assumption. 
   + destruct H as [Heq | Hsub Hwf];
     [ rewrite <- Heq; apply (@wf_prove Ga A H0); auto 
     | apply Hsub].

 - intros; apply weakening_stoup_gen with Ga; auto.
   + apply indeed_sub_ctx_is; assumption.
   + destruct H as [Heq | Hsub Hwf];
     [ rewrite <- Heq; apply (@wf_prove_stoup Ga A C H0); auto
     | apply Hsub].
Defined.     

Lemma weakening_atom : forall (Ga Ga': context),
    sub_ctx Ga Ga' ->
    forall (P : predicate * wf_trm),
    wf_atoms Ga P ->
    wf_atoms Ga' P.
Proof. 
  intros; apply weakening with Ga; 
  [ apply indeed_sub_ctx_is
  | destruct H as [Heq | Hsub ?];
    [ rewrite <- Heq; unfold wf_atoms in H0;
      apply (@wf_prove Ga (Atom (fst P, proj1_sig (snd P))))
     |apply Hsub]
    | ]; auto.
Defined.

(** Reflexivity of sub-context relation. *)

Lemma sub_ctx_refl : forall Ga, sub_ctx Ga Ga.
Proof.
  unfold sub_ctx; intuition. 
Qed.

(** Transitivity of sub-context relation. *)

Lemma sub_ctx_trans : forall Ga Ga0 Ga1, 
  sub_ctx Ga Ga0 -> sub_ctx Ga0 Ga1 -> sub_ctx Ga Ga1.
Proof.
  unfold sub_ctx; intros Ga Ga0 Ga1 H H0.
  destruct H as [Heq | Hneq]; destruct H0 as [H0eq | H0neq]; 
  subst; intuition. 
Qed.

(** Universal model's domain is the set of well-formed terms. *)

Canonical Structure Universal := @Build_Kripke
  context
  sub_ctx
  sub_ctx_refl
  sub_ctx_trans
  wf_trm
  (wf_Cst)
  (wf_App)
  (fun Ga Pt => Ga |- Atom (fst Pt, proj1_sig (snd Pt)))
  weakening_atom.

(** In the universal model, substitution plays the role of
   semantic interpretation. *)

Lemma psem_tsubst : forall (t:trm) (rho eta:assoc wf_trm),
  sub_name (ov_t t) (dom eta) -> sub_name (op_t t) (dom rho) -> 
  >> (psem Universal t rho eta) = (t [[>>> rho , >>> eta]]).
Proof.
  induction t; try tauto; intros.  
  - simpl in *. apply In_incl in H. 
    generalize H; intro H1; rewrite fn_pro_dom_eq in H1.
    apply lookup_exists in H1; apply lookup_exists in H. 
    inversion H1; inversion H. rewrite H2, H3; auto. 
    apply fn_pro_eq_val with (eta:=eta) (n:=n); auto. 
  - simpl in *. apply In_incl in H0. 
    generalize H0; intro H1; rewrite fn_pro_dom_eq in H1.
    apply lookup_exists in H0; apply lookup_exists in H1;
    inversion H0; inversion H1.
    rewrite H2, H3; apply fn_pro_eq_val with (eta:=rho) (n:=n); auto.
  - simpl in *; f_equal; auto; [apply IHt1| apply IHt2]; auto with datatypes.  
Defined.

  (** ** Universal Completeness *)

Theorem univ_cpltness : forall (A:formula) (Ga:context) (rho eta:assoc wf_trm),
    sub_name (ov A) (dom eta) -> 
    sub_name (op A) (dom rho)->
    wf_c Ga ->
    Ga ||- A {[rho, eta]} ->
    Ga |- (A [>>> rho, >>> eta])

    with univ_cpltness_stoup : forall (A:formula) (Ga:context) (rho eta:assoc wf_trm),
    sub_name (ov A) (dom eta) ->
    sub_name (op A) (dom rho) ->
    wf_c Ga ->
    (forall C Ga', sub_ctx Ga Ga' ->
                   (Ga' ;; (A [>>> rho, >>> eta]) |- C )-> (Ga' |- C )) ->
    Ga ||- A {[rho, eta]}.
Proof.
  - {destruct A as [ [p t] | | ]; simpl; intros.
     - rewrite <- psem_tsubst; assumption.
     - generalize (incl_appL H); intro.
       generalize (incl_appR H); intro; clear H.
       apply ProofImplyR.
       apply univ_cpltness; auto with datatypes.
       
       + apply wCons; auto.
         apply empty_ov_wf, wf_trm_subst_wf; auto with datatypes.
       + apply X; auto. 
         * unfold sub_ctx; right. split;
                                  [ idtac 
                                  | apply wCons; auto; 
                                    apply empty_ov_wf, wf_trm_subst_wf]; 
           auto with datatypes.
         * apply univ_cpltness_stoup; auto with datatypes;
            [apply wCons; auto;
             apply empty_ov_wf, wf_trm_subst_wf; auto
            | intros; apply ProofCont with  (A1 [>>> rho, >>> eta]); 
              unfold sub_ctx in H]; auto.
             destruct H; [rewrite <- H | apply H]; auto with datatypes. 
          
     - set (a:= new (op_c ((Forall n (A [>>> rho, >>> eta ^- n])) :: Ga))).
              apply ProofForallR with a; intros.
              apply new_fresh.  unfold open; rewrite open_eq_subst. 
       + rewrite <- fn_pro_cons;
         apply univ_cpltness; auto.
         * simpl; apply (@incl_rm_2 name (name_Dec)); auto.
         * apply X; apply sub_ctx_refl. 
       + apply wPar. 
          }
  - { destruct A as [ [p t] | |]; simpl; intros.
      - apply H2; auto using sub_ctx_refl. 
        rewrite psem_tsubst; auto. 
        apply ProofAxiom; auto; apply wAtom; apply wf_t_sub_ov; auto.
      - apply univ_cpltness_stoup; auto with datatypes.
        + apply sub_ctx_wf_c with Ga; auto. 
        + intros; apply H2; auto. 
          * apply wle_trans with w'; auto. 
          * { apply ProofImplyL ; auto. 
               apply univ_cpltness; auto with datatypes. 
              - apply sub_ctx_wf_c with Ga; auto. apply sub_ctx_trans with w'; auto.  
              - apply mono_force with w'; auto.
}
      - apply univ_cpltness_stoup; auto. apply incl_rm_2; auto.
       * apply sub_ctx_wf_c with Ga; auto. 
       * intros; apply H2; auto. 
        + apply wle_trans with w'; auto.
        + destruct d as [m wf]; apply ProofForallL with m; auto. 
          unfold open; rewrite open_eq_subst; auto. 
}
Defined.


Notation cl_a := (fun (A: formula) => sub_name (op A) nil).

Notation cl_c :=
  (fun (Ga:context) => forall A, A @ Ga -> sub_name (op A) nil).

Theorem univ_cpltness' : 
  forall A (Ga: context),
    cl_a A ->
    wf A -> wf_c Ga ->
    Ga ||- A {[nil, nil]} ->
    Ga |- A.
Proof.
  intros; rewrite nil_subst. 
  rewrite <- fn_pro_nil; apply univ_cpltness; auto; 
  rewrite wf_ov_empty; auto with datatypes.
Defined.

Theorem univ_cpltness_stoup' : forall A (Ga: context),
    cl_a A ->
    wf A -> wf_c Ga ->
    (forall C Ga', sub_ctx Ga Ga' -> Ga' ;; A |- C -> Ga' |- C) ->
    Ga ||- A {[nil, nil]}.
Proof. 
  intros.
  apply univ_cpltness_stoup; auto;
  [ rewrite wf_ov_empty; auto with datatypes
  | intros; apply H2; auto; rewrite nil_subst with A; rewrite <- fn_pro_nil; auto].
Defined.

Lemma univ_cpltness_axiom : forall A (Ga : context),
    cl_a A -> wf A -> wf_c Ga ->
    A @ Ga ->
    Ga ||- A {[nil, nil]}.
Proof.
  induction Ga as [ | B Ga]; try contradiction; simpl; intros. 
  apply univ_cpltness_stoup; auto. 
  - rewrite wf_ov_empty; auto. 
  - intros. apply ProofCont with (A [>>> nil, >>> nil]) ; auto. 
    rewrite! fn_pro_nil; rewrite <- nil_subst.
    destruct H3; intuition auto with datatypes; rewrite <- H3; auto with datatypes.
    + rewrite <- H5; auto with datatypes.
Defined.

Lemma universal_force_context : forall (Ga: worlds Universal) De,
    cl_c De -> wf_c De ->
    sub_ctx De Ga -> 
    list_forall (fun B => Ga ||- B {[nil, nil]}) De.
Proof.
  induction De as [|A]; intros; constructor.
  - apply univ_cpltness_axiom; auto using in_eq; auto.
    + inversion H0; subst; auto. 
    + unfold sub_ctx in H1; intuition; subst; auto. 
    + unfold sub_ctx in H1; destruct H1; subst. 
      * apply in_eq.  
      * apply H1; auto with datatypes. 
  - apply IHDe; auto with datatypes. 
    * inversion H0; auto. 
    * unfold sub_ctx; destruct H1; subst; auto with datatypes. 
      right; destruct H1; eauto using incl_cons_1. 
Defined.

Theorem completeness : forall (Ga: context) A,
    cl_a A ->
    cl_c Ga ->
    wf A -> wf_c Ga ->
    (forall K (w: worlds K),
        (list_forall (fun B => w ||- B {[nil, nil]}) Ga) -> w ||- A {[nil, nil]}) ->
    Ga |- A.
Proof.
  intros; apply univ_cpltness'; auto. 
  apply X. apply (universal_force_context H0 H2 (sub_ctx_refl Ga)). 
Defined.

(** ** Cut-admissibility *)

Lemma cut_ad : 
  forall Ga A B,
    cl_c Ga -> cl_a B-> 
    Ga ;; A |- B -> Ga |- A -> Ga |- B.
Proof.
  intros.
  apply completeness; auto. 
  - destruct (wf_prove_stoup H1) as [H3 H4]; destruct H4; auto.
  - apply (wf_prove H2).
  - intros. apply soundness_stoup with (Ga:=Ga) (A:=A) (C:=B); auto. 
    apply soundness with Ga; auto.
Defined.