(** Renaming of Constants *)

Set Implicit Arguments.

Require Import environments.
Require Import variable_sets.
Require Import extralibrary.
Require Import substitution.
Require Import sublist.
Require Import language_syntax.


Definition substitution := assoc trm.

Notation renaming := (assoc name).

Fixpoint renaming_emb (rho: renaming) : substitution :=
  match rho with
  | nil            => nil
  | (x, a) :: rho' => (x, Par a) :: renaming_emb rho'
  end.

(** [McKinna & Pollak] and [Leroy] used [simultaneous] parameter swapping
   in order to show the equivalence between different styles of quantification.

   We will use simultaneous [renaming] of constants.
   Note that [renaming] is easier to deal with. *)

(** [renaming] is an arbitrary funtion from [Cst] to [Cst]. *)


Fixpoint rename_t (eta : renaming) (t : trm) : trm := 
  match t with
    | Var y    => Var y
    | Par x   => match eta ^^ x with 
                     | Some y => Par y
                     | None => Par x
                 end
    | Cst c   => Cst c                
    | App f t1 t2 => App f (rename_t eta t1)  (rename_t eta t2)
  end.

Fixpoint rename (eta : renaming) (A : formula) : formula := 
  match A with
    | Atom (p,t) => Atom (p, rename_t eta t )
    | B --> C    => (rename eta B) --> (rename eta C)
    | Forall y B => Forall y (rename eta B)
  end.

Fixpoint rename_c (eta : renaming) (Ga : context) : context := 
  match Ga with
    | nil => nil
    | A :: Ga0 => rename eta A :: rename_c eta Ga0
  end.

(** Renaming associations *)

Fixpoint rename_a (eta : renaming) (rho : assoc trm) :=
  match rho with
    | nil           => nil
    | (x,t) :: rho0 => (x, rename_t eta t) :: rename_a eta rho0
  end.


(** Renaming is a special case of substitution *)

Lemma renaming_Some_subst : 
  forall rho m a,
    rho ^^ m = Some a -> renaming_emb rho ^^ m = Some (Par a).
Proof.
  induction rho as [| [n t] rho]; auto; simpl; intros. 
  - discriminate.
  - case_name; solve[auto | congruence].
Defined.

Lemma renaming_None_subst : 
  forall rho m,
    rho ^^ m = None -> renaming_emb rho ^^ m = None.
Proof.
  induction rho as [| [n t] rho]; auto; simpl; intros. 
  - case_name; solve[auto | congruence].
Defined.

Implicit Arguments renaming_Some_subst [rho m a].
Implicit Arguments renaming_None_subst [rho m].


(** No renaming means no change  *)

Lemma rename_t_nil : forall (t : trm),
  rename_t nil t = t.
Proof.
  induction t; simpl; f_equal; auto.
Qed.

Lemma rename_nil : forall (A:formula),
  rename nil A = A.
Proof.
  induction A ; [destruct a| |]; simpl; 
  try f_equal; try rewrite rename_t_nil; auto.
Qed.

Lemma rename_c_nil : forall (Ga:context),
  rename_c nil Ga = Ga.
Proof.
  induction Ga; simpl; f_equal; auto using rename_nil.
Qed.

(** Compatibility of renaming with predicates and operators. *)

Lemma rename_In_ctx : 
  forall A Ga eta,
    A @ Ga ->
    (rename eta A) @ (rename_c eta Ga).
Proof.
  induction Ga; simpl; intros;
  repeat case_lang; auto; intuition.
  left; subst; auto.
Qed.

Lemma In_dom_sig_rename : 
  forall (x: name) (rho: assoc trm)(eta:renaming) t,
    rho ^^ x = Some t ->
    (rename_a eta rho) ^^ x = Some (rename_t eta t).
Proof.
  induction rho as [ | [y b] ]; simpl; intros;
    [ discriminate
      | case_name; inversion H; subst; auto
    ].
Qed.

Lemma In_dom_sig_none_rename :
  forall x rho eta,
    rho ^^ x = None ->
    rename_a eta rho ^^ x = None.
Proof.
  induction rho as [| [y b] rho ]; auto; simpl; intros.
  case_name; 
  [discriminate
  | rewrite IHrho; auto].
Qed.

Hint Resolve rename_In_ctx In_dom_sig_rename.
Hint Rewrite rename_In_ctx In_dom_sig_rename.

(** Renaming and substitution *)

Lemma rename_comm_assoc_rm :
  forall zeta eta n,
 rename_a zeta (eta ^- n) = (rename_a zeta eta) ^- n.
Proof.
  induction eta as [| [m t]  eta]; auto; simpl; intros. 
  case_name; auto. 
  rewrite <- IHeta; auto.  
Qed.

Lemma rename_neq_rm : 
  forall zeta eta n m,
    n <> m -> 
    (rename_a zeta (eta ^- m)) ^^ n = (rename_a zeta eta) ^^ n. 
Proof.
  induction eta as [| [m t] eta]; auto; simpl; intros.
  do 2 case_name; subst; auto; try tauto.
  - rewrite <- rm_change; auto; rewrite rename_comm_assoc_rm.
    rewrite <- env_rm_neq; auto;simpl; case_name; tauto. 
  - simpl; case_name; subst; auto; tauto.
Qed.

Lemma rename_sub_t :
  forall (t : trm) (rho : assoc trm) (zeta : renaming) (n : name),
    rename_t zeta (t [[nil, (n, Var n) :: rho]]) =
    ((rename_t zeta t) [[nil, rename_a zeta rho ^- n]]).
Proof. 
  induction t; auto; intros; rewrite <- rename_comm_assoc_rm; simpl.
  - case_name; subst; auto. 
    + rewrite rename_comm_assoc_rm, env_rm_self; auto.
    + rewrite rename_neq_rm; auto.
      pose proof (In_dom_sig_rename n rho zeta);
      pose proof (In_dom_sig_none_rename n rho zeta).
      case_Env;
      [ rewrite H with t
      | rewrite H0] ; auto. 
  - case_Env; auto.
  - rewrite IHt1, IHt2, rename_comm_assoc_rm; auto.
Qed.

Lemma  rename_sub :
  forall (A : formula) (zeta : renaming) (eta : assoc trm) (n : name),
    rename zeta (A [nil, (n, Var n) :: eta]) =
    (rename zeta A [nil, rename_a zeta eta ^- n]).
Proof.
  induction A as [ [p t] | |]; auto; simpl; intros.
  - do 2 f_equal; rewrite <- rename_sub_t; auto.
  - rewrite <- IHA1, IHA2; auto.
  - f_equal; case_name; auto. 
    * rewrite <- rm_rm, <- IHA, subst_rm; auto.
    * rewrite <- rm_neq_rm; auto; rewrite <- rename_comm_assoc_rm; rewrite IHA; auto.
Qed.


Lemma rename_subst_t :
  forall (t : trm) (eta : assoc trm) (zeta : renaming),
    sub_name (ov_t t) (dom eta) ->
    rename_t zeta (t [[nil , eta]]) = (rename_t zeta t [[nil , rename_a zeta eta]]).
Proof.
  induction t; auto; simpl; intros. 
  - apply incl_cons_0 in H.  
    apply lookup_exists in H; destruct H as [x H0]. 
    rewrite H0. rewrite In_dom_sig_rename with (t:=x); auto.
  - case_Env; auto. 
  - f_equal; [rewrite IHt1 | rewrite IHt2]; auto with datatypes. 
Qed.

Lemma rename_subst :
  forall (A : formula) (eta : assoc trm) (zeta : renaming),
    sub_name (ov A) (dom eta) ->
    rename zeta (A [nil, eta]) = (rename zeta A [nil, rename_a zeta eta]).
Proof.
induction A as [ [p t] | |] ; simpl; intros.
 - rewrite rename_subst_t; auto.
 - rewrite IHA1, IHA2; auto with datatypes.
 - f_equal; rewrite subst_rm, rename_sub; auto. 
Qed.

(** ** Renaming and freshness *)

Lemma rename_t_fresh : forall (t:trm) eta a a0,
  a # op_t t ->
  rename_t ((a,a0) :: eta) t = rename_t eta t.
Proof.
  induction t; simpl; intros; auto;
  [ repeat case_name; try congruence; elim H 
  | f_equal; [apply IHt1 | apply IHt2] ]; auto with datatypes.
Qed.

Lemma rename_fresh : forall (A:formula) eta a a0,
  a # op A ->
  rename ((a, a0) :: eta) A = rename eta A.
Proof.
  induction A as [ [p t] | | ]; simpl; intros. 
  - do 2 f_equal; apply rename_t_fresh; auto.
  - rewrite IHA1, IHA2; auto with datatypes.
  - f_equal; auto.
Qed.

Lemma rename_c_fresh : forall Ga eta a a0,
  a # op_c Ga ->
  rename_c ((a, a0) :: eta) Ga = rename_c eta Ga.
Proof.
  induction Ga as [ | A Ga0]; simpl; intros; auto.
  destruct_notIn; simpl; f_equal; 
  [rewrite rename_fresh | rewrite IHGa0]; auto with datatypes.
Qed.

(** Function values *)

Definition fun_value (eta : renaming) (a:name) : name :=
  match eta ^^ a with
    | Some a0 => a0
    | None   => a
  end.

Notation " eta ** a " := (fun_value eta a) (at level 60).

Lemma trm_rename_fresh : forall (t:trm) eta a a0,
  a0 # op_t t ->
  a0 # dom eta ->
  a0 # img eta ->
  rename_t ((a0, eta ** a) :: eta) (rename_t ((a, a0) :: nil) t) = 
  rename_t eta t.
Proof.
  induction t; simpl; intros; destruct_notIn; auto. 
  - repeat case_name; simpl; firstorder. 
    + rewrite (notIn_dom_None _ _ H0); simpl; 
          case_name; unfold fun_value;
            case_Env; intuition. 
    + case_name; intuition. 
  - f_equal; [rewrite IHt1 | rewrite IHt2]; auto with datatypes.
Qed.

Lemma fml_rename_fresh : forall (A:formula) eta a a0,
  a0 # op A -> 
  a0 # dom eta ->
  a0 # img eta ->
  rename ((a0, eta ** a) :: eta) (rename ((a, a0) :: nil) A) = 
  rename eta A.
Proof.
   induction A as [ [p t] | |]; simpl; intros;
   [ do 2 f_equal; auto using trm_rename_fresh
   | destruct_notIn; f_equal; auto with datatypes 
   | f_equal; auto with datatypes].
Qed.

Lemma context_rename_fresh : forall De eta a a0,
  a0 # op_c De ->
  a0 # dom eta ->
  a0 # img eta ->
  rename_c ((a0, eta ** a) :: eta) (rename_c ((a, a0) :: nil) De) =
  rename_c eta De.
Proof.
  induction De as [ | A]; simpl; intros; auto;
  destruct_notIn; f_equal; [rewrite fml_rename_fresh | rewrite IHDe]; auto with datatypes.    
Qed.

Lemma notIn_op_rename_t : forall (t : trm) eta a a0,
  a0 # img eta ->
  a <> a0 ->
  a0 # op_t (rename_t ((a0, eta ** a) :: eta) t).
Proof.
  induction t; simpl; intros; auto;
  [ case_name; auto; unfold fun_value; case_Env;
    [ intro F; simpl in F; destruct F; auto; rewrite H2 in *; apply H;
      elim (@env_dom_notIn_img _ _ _ eta a0 H a); auto;
      apply (lookup_inv) in H1; auto
    | simpl; intuition
    | intro F; simpl in F; destruct F; auto; rewrite H2 in H1;
      inversion H1; apply lookup_inv in H4; 
      elim (@env_dom_notIn_img _ _ _ eta a0 H n); auto
    | simpl; intuition]
    | solve_notIn].
Defined. 

Lemma notIn_op_rename : forall (A : formula ) eta a a0,
  a0 # img eta ->
  a <> a0 ->
  a0 # op (rename ((a0, eta ** a) :: eta) A).
Proof.
  induction A as [ [p t] | |]; simpl; intros;
    [ auto using notIn_op_rename_t
      | solve_notIn
      | auto
    ].
Qed.

Lemma notIn_op_rename_c : forall (Ga : context) eta a a0,
  a0 # img eta ->
  a <> a0 ->
  a0 # op_c (rename_c ((a0, eta ** a) :: eta) Ga).
Proof.
  induction Ga as [ | A Ga0]; simpl; intros;
    auto using notIn_app_2, notIn_op_rename.
Qed.

Lemma sub_ctx_fresh_par :
  forall Ga Ga0 a a0,
    sub_ctx_pre Ga Ga0 ->
    a # op_c Ga ->
    sub_ctx_pre Ga (rename_c ((a, a0) :: nil) Ga0).
Proof.
  unfold sublist; intros.
  assert (a1 = rename ((a, a0):: nil) a1); 
  [ rewrite rename_fresh ; 
    [ symmetry; apply rename_nil 
    | contradict H0; 
      cut (sub_name (op a1) (op_c Ga)); auto with datatypes]
  | rewrite H2; apply rename_In_ctx; auto].
 Defined.


(* a bunch of definitions about subst/open belongs language_syntax *)

Lemma par_sub_par : 
  forall rho m u,
    renaming_emb rho ^^ m = Some u -> exists a, u = Par a.
Proof. 
  induction rho as [| [n t] rho]; auto; simpl; intros. 
  - discriminate.
  - destruct (eq_name m n); auto. 
    + exists t; inversion H; auto. 
    + apply IHrho with m; auto. 
Defined.


Lemma open_subst_t : 
  forall u (eta: renaming) n t, 
    open_t (u [[nil , renaming_emb eta ^- n]]) n t 
    = (u [[ nil , (n, t) :: renaming_emb eta]]).
Proof.
  unfold open_t; induction u as [ m | m | m | ]; simpl; intros; auto.
  - destruct (eq_name m n); subst; auto.
    + rewrite env_rm_self; simpl; destruct (eq_name n n); intuition. 
    + rewrite <- env_rm_neq; auto.
      pose proof (par_sub_par eta m).
      destruct (renaming_emb eta ^^ m). 
      * { pose proof (H t0).
          assert (exists c, t0 = Par c).
          - apply H0; auto.
          - destruct H1; rewrite H1; auto.   
        }
      * simpl; auto. 
        case_name; firstorder.
  - f_equal; auto. 
Defined.

Lemma emb_rm : 
  forall rho m,
    renaming_emb (rho ^- m) = renaming_emb rho ^- m.
Proof.
  induction rho as [| [n n'] rho]; auto; simpl; intros.
  case_name; auto.
  simpl; f_equal; auto. 
Defined.

Lemma open_subst :
  forall A eta n t, 
    open (A [nil, renaming_emb eta ^- n]) n t 
    = ( A [nil , (n, t) :: renaming_emb eta]).
Proof.
  unfold open; induction A as [ [P u]| | m ]; simpl; intros; f_equal; auto.
  - f_equal;
    apply open_subst_t; auto. 
  - case_name. 
    * subst; rewrite <- nil_subst, <- rm_rm; auto. 
    * rewrite rm_neq_rm, <- emb_rm; auto. 
Defined.


(** Well-formedness of provable formulas *)

Lemma ov_wf_subst_t : 
  forall t (rho : renaming),
    sub_name (ov_t t) (dom rho) ->
    wf_t (t [[nil, renaming_emb rho]]).
Proof. 
  induction t; auto; intros; simpl.
  - apply In_incl in H. 
    assert (n @ dom (renaming_emb rho)).
    + induction rho as [ | [x b] rho]; auto; simpl.
      change (dom ((x, b) :: rho)) with (x :: dom rho) in H;
    change (dom ((x, Par b) :: renaming_emb rho)) 
    with ( x :: dom (renaming_emb rho)).
      simpl in H; destruct H as [HL | HR]; [intuition | ]. 
      right; apply IHrho; auto. 
    + case_Env.
      * destruct (par_sub_par rho n H1); subst.
        apply wPar.
      * destruct (@env_lookup_notIn name _ trm n (renaming_emb rho) H1); auto.
  - apply wPar.
  - apply wCst.
  - simpl in *; constructor; auto with datatypes.
Qed. 

Lemma ov_wf_subst : forall A (rho : renaming), 
  sub_name (ov A) (dom rho) ->
  wf (A [nil, renaming_emb rho]).
Proof.      
  induction A as [ [p t] | |]; simpl; intros rho Hsub.
  - apply wAtom; apply ov_wf_subst_t; auto. 
  - apply wImply; auto with datatypes.
  - set (a:=  (new (op (A [nil, renaming_emb rho ^- n])))).
    apply wForall with (c := a). 
    apply new_fresh.
    + rewrite open_subst; 
        change ((n, Par a) :: renaming_emb rho) 
              with (renaming_emb ((n, a) :: rho)); 
        apply IHA; simpl in *; apply incl_rm_2; auto.
Defined.

Lemma empty_ov_wf_t : forall t, ov_t t = nil -> wf_t t.
Proof.
  induction t; simpl; intros.
  - discriminate.
  - apply wPar.
  - apply wCst.
  - apply wApp;
    simpl;
     try apply IHt1;
      apply app_eq_nil in H; firstorder.
Qed.


Lemma empty_ov_wf : forall A, ov A = nil -> wf A.
Proof.
  intros A Hov. 
  cut (A = (A [nil, renaming_emb nil])).
  - intro H; rewrite H.
    apply ov_wf_subst; rewrite Hov; auto with datatypes.
  - simpl; auto using nil_subst with datatypes. 
Defined. 

Hint Resolve ov_wf_subst.

Lemma open_forall_wf : forall A x u, wf (open A x u) -> wf (Forall x A).
Proof.
  intros A x u Hwf. 
  apply empty_ov_wf; apply wf_ov_empty in Hwf.
  apply incl_nil; rewrite <- Hwf; auto using sub_q_open.
Defined.

Hint Resolve sub_q_open open_forall_wf.
Hint Rewrite sub_q_open.


(** Well-Formedness and Renaming *)

Lemma rename_wf_t_eq : forall t eta,
  ov_t (rename_t eta t) = ov_t t.
Proof.
  induction t; auto; simpl; intros;
  [ case_Env
  | rewrite IHt1, IHt2]; auto.
Defined.

Hint Resolve rename_wf_t_eq. Hint Rewrite rename_wf_t_eq.
 
Lemma rename_wf_t : forall t eta,
  wf_t t -> wf_t (rename_t eta t).
Proof.
  intros; 
  apply empty_ov_wf_t.
  apply wf_t_ov_empty in H.
  rewrite <- H; auto.
Defined. 

Hint Resolve rename_wf_t. 

Lemma rename_wf_eq : forall A eta,
  ov (rename eta A) = ov A.
Proof.
  induction A; try destruct a; simpl; auto; intros;
  [ rewrite IHA1, IHA2| rewrite IHA] ; auto.
Defined.
Hint Resolve rename_wf_eq. Hint Rewrite rename_wf_eq.

Lemma rename_wf_f : forall A eta,
  wf A -> wf (rename eta A).
Proof.
  intros; apply wf_ov_empty in H; apply empty_ov_wf; rewrite <- H; auto.
Defined.

Hint Resolve rename_wf_f.

Lemma rename_wf_c : forall De eta,
  wf_c De -> wf_c (rename_c eta De).
Proof.
  induction De; simpl; auto; intros.
  inversion H; apply wCons; auto using IHDe. 
Defined.

Hint Resolve rename_wf_c.
