(** * Substitution *)

Set Implicit Arguments.

Require Import decidable_In.
Require Import environments.
Require Import sublist.
Require Import language_syntax.
Require Import extralibrary.
Require Export fresh_name.
Require Import variable_sets.


(** Defining substitution requires special consideration for two reasons.

   - First, the relationship between syntax and semantics such as soundness
     and completeness should be realized in a natural way.

     For this purpose, we find that simultaneous substitution provides
     a convenient way in establishing the relationship.

   - Second, because of the trace part, it is not clear to which class
     the resulting expression of a substitution should belong.

     Thus we decided to state clearly the type of the resulting expression
     in the definition of substitution.


   - Notice the special role of the [Var] case because it needs relocation.

   - No alpha-conversion is used.

   - Since there are no parameters, substitution only for variables are necessary. *)

(** The type of simultaneous substitutions =: [name -> trm].
    Instead of using funntion type, lists of pairs of names and terms can be
    used for the simultaneous substitutions, but this is less handy. *)


(** *** Substitution functions *)

(** The type of simultaneous substitutions =: [name -> trm].
    Instead of using funntion type, lists of pairs of names and terms can be
    used for the simultaneous substitutions, but this is less handy. *)

Notation assoc := (env name).

Reserved Notation "t [[ rho , eta ]]" (at level 69).
Reserved Notation "A [ rho , eta ]" (at level 69).
Reserved Notation "Ga [! rho , eta !]" (at level 69).

Fixpoint subst_t (t : trm) (rho eta:assoc trm) {struct t} : trm :=
  match t with
    | Var y       => match eta ^^ y with
                     | Some u => u
                     | None   => Var y
                     end
    | Par x       => match rho ^^ x with
                     | Some u => u
                     | None   => Par x
                     end
    | Cst c       => Cst c
    | App f t1 t2 => App f (t1 [[rho, eta]]) (t2 [[rho, eta]])
  end
  where "t [[ rho , eta ]]" := (subst_t t rho eta).

Notation "t [[ rho , eta ]]" := (subst_t t rho eta) (at level 69).

Fixpoint subst (A : formula) (rho eta: assoc trm) {struct A} : formula :=
  match A with
    | Atom (P,t) => Atom (P, t [[ rho , eta ]] )
    | B --> C    => (B [ rho, eta ]) --> (C [ rho, eta ])
    | Forall y B => Forall y (B [ rho , eta ^- y])
  end
  where "A [ rho , eta ]" := (subst A rho eta).

Notation "A [ rho , eta ]" := (subst A rho eta) (at level 69).

Fixpoint subst_c (Ga : list formula) (rho eta: assoc trm) {struct Ga} : list formula :=
  match Ga with
    | nil => nil
    | A :: Ga' => (A [ rho , eta]) :: (Ga' [! rho , eta !])
  end
  where "Ga' [! rho , eta !]" := (subst_c Ga' rho eta).

Notation "Ga [! rho , eta !]" := (subst_c Ga rho eta) (at level 69).

(* No substitution means no change  *)

Lemma nil_subst_t : 
  forall t, t = (t [[ nil, nil ]] ).
Proof.
induction t; simpl; auto; f_equal; auto.
Defined. 

Lemma nil_subst : forall A,
 A = (A [nil , nil]).
Proof.
- induction A as [ [p t] | |]; simpl; intros.
  + rewrite <- nil_subst_t; auto.
  + f_equal; auto.
  + rewrite <- IHA; auto.
Defined.

Hint Resolve nil_subst_t nil_subst.
Hint Rewrite nil_subst_t nil_subst.


(** Opening a term: replacing a variable by a term *)

Definition open_t t x u := t [[nil , (x,u)::nil ]].

Definition open A x u := A [ nil , (x,u)::nil ].


(** ** Well-formedness *)

(** Well-formed expressions are expressions without any free occurrence of letters. *)

Inductive wf_t : trm -> Prop :=
| wCst : forall x, wf_t (Cst x)
| wPar : forall x, wf_t (Par x)
| wApp : forall f t1 t2, wf_t t1 -> wf_t t2 -> wf_t (App f t1 t2).

Inductive wf : formula -> Prop :=
| wAtom  : forall P t, wf_t t -> wf (Atom (P,t))
| wImply : forall A1 A2, wf A1 -> wf A2 -> wf (A1 --> A2)
| wForall: forall x c A, c # op A -> wf (open A x (Par c)) -> wf (Forall x A).

(** A list of well-formed formulas is called [context].
    That condition is expressed by ok-ness. *)

Inductive wf_c : list formula -> Prop :=
| wNil : wf_c nil
| wCons : forall Ga A, wf_c Ga -> wf A -> wf_c (A :: Ga).


(* well-formedness and occurence of variables *)


Lemma rm_ov_t_eq : forall t c x,
 (ov_t t) \ x = ov_t (t [[nil, (x, Par c) :: nil]]).
Proof.
  induction t; simpl; intros; auto with datatypes.
  - repeat case_name; auto; 
    elim n0; auto.
  - rewrite <- IHt1, <- IHt2; auto with datatypes. rewrite <- (@rm_app name (name_Dec)); auto.
Defined.

Hint Resolve rm_ov_t_eq.
Hint Rewrite rm_ov_t_eq.

Lemma rm_ov_eq : forall A c x,
  (ov A) \ x = ov (A [nil, (x, Par c) :: nil]).
Proof.
  induction A as [ [p t] | | ]; intros; simpl; auto with datatypes.
  - rewrite <- IHA1, <- IHA2; rewrite (@rm_app name (name_Dec)); auto. 
  - case_name; [rewrite <- nil_subst | rewrite <- IHA]; auto. 
    + rewrite (@rm_rm_eq_rm name (name_Dec)); auto. 
    + rewrite (@rm_neq_eq name (name_Dec)); auto. 
Defined.

Hint Resolve rm_ov_eq.
Hint Rewrite rm_ov_eq.


Lemma wf_t_ov_empty : forall t, wf_t t -> ov_t t = nil.
Proof.
  induction t; simpl; intros; inversion H; auto;
  rewrite IHt1, IHt2; auto.
Defined.

Hint Resolve wf_t_ov_empty.
Hint Rewrite wf_t_ov_empty.

Lemma wf_ov_empty : forall A, wf A -> ov A = nil.
Proof.
  induction 1; simpl; auto with datatypes. 
  - rewrite IHwf1, IHwf2; auto.
  - rewrite <- IHwf;
    rewrite rm_ov_eq with (c:=c); auto. 
Defined. 

Hint Resolve wf_ov_empty.
Hint Rewrite wf_ov_empty.

Lemma sub_q_open : 
  forall A x u,
    sub_name (ov (Forall x A)) (ov (open A x u)).
Proof. 
  induction A as [ [ p t ] | |] ; simpl; intros.
  - induction t ; auto; simpl.
    + do 2 case_name; auto with datatypes;
      congruence.
    + rewrite <-  (@rm_app name (name_Dec)); auto with datatypes. 
  - simpl in *; unfold open in * |-; auto with datatypes. 
    rewrite <- (@rm_app name (name_Dec)); auto with datatypes. 
  - case_name; 
    [ rewrite <- nil_subst, (@rm_rm_eq_rm name (name_Dec)); auto with datatypes 
    | rewrite (@rm_neq_eq name (name_Dec)); apply (@incl_rm_1 name (name_Dec)), IHA].
Defined.

Lemma wf_c_wf : forall Ga A,
  A @ Ga -> (wf_c Ga) -> wf A.
Proof.
  induction Ga as [| B Ga]; intros; auto. 
  - inversion H. 
  - simpl in H; destruct (fml_dec A B); inversion H0; subst; auto.
    destruct H as [H | H]; apply IHGa; auto. elim n; auto.    
Defined.

Hint Resolve wf_c_wf. 

Lemma wf_fml_sub_name : forall x A, 
  wf (Forall x A) -> sub_name (ov A) (x :: nil).
Proof.
  intros; destruct (wf_ov_empty H); simpl; apply incl_rm_2, incl_refl. 
Defined.

(** The rule [wForall] deserves attention. Since when a quantification 
  is just eliminated from a formula, a variable becomes suddenly
  unbound. This is the reason why just structural induction does not
  work.  

  There are several ways to deal with this problem. Here we use the
  technique of opening and replace a binder with _one_ fresh
  constant. Many discussions have been made about the style of
  opening. People used to say that the style we use is not so flexible
  enough for doing formalization work althogh the style we use here is
  most close to the pencil-and-paper practice of binding. 

  However, we will show that one could work smoothly with the style of
  dealing with opening we use here. *)

(* the following are used in completeness_and_NBE *)
(* projections related with well-formedness  *)

Notation wf_trm := (@sig trm wf_t).

Definition wf_Cst (n:name) : wf_trm := 
  exist wf_t (Cst n)(wCst n).

Definition wf_Par (n:name) : wf_trm := 
  exist wf_t (Par n)(wPar n).

Definition wf_App (f:function)(t u: wf_trm) : wf_trm :=
  exist wf_t
        (App f (proj1_sig t) (proj1_sig u)) 
        (wApp f (proj2_sig t) (proj2_sig u)).

(** ** Relation wrt contexts and formulas *)

(** The sub-context relation plays an important role in dealing with
  provability and Kripke semantics. However, we have to focus on
  well-formed contexts.

  First the sub-context relation [sub_ctx_pre] is defined for all
  contexts then we will focus on well-formed contexts when [sub_ctx]
  is defined. *)

(** * sub_ctx relation : this is used in completeness  *)

Definition sub_ctx (Ga Ga' : context) : Prop :=
  (Ga = Ga') \/( (sub_ctx_pre Ga Ga' /\ wf_c Ga')).

Lemma indeed_sub_ctx_is : 
  forall Ga Ga',
    sub_ctx Ga Ga' ->
    sub_ctx_pre Ga Ga'.
Proof.
  intros. destruct H as [Heq | Hsub Hwf];
  [ rewrite Heq; auto with datatypes 
  | apply Hsub].
Defined.

Lemma sub_ctx_wf_c : forall Ga Ga0,
    sub_ctx Ga Ga0 -> wf_c Ga -> wf_c Ga0.
Proof.
  intros; unfold sub_ctx in H. 
  destruct H as [Heq | Hneq]; subst; intuition. 
Defined.

Notation ">>" := (proj1_sig).
Notation "<<" := (exist wf_t).


Fixpoint fn_pro (eta : assoc wf_trm) :=
  match eta with
    nil => nil
  | (n, t) :: eta' => (n, >> t) :: (fn_pro eta')
  end.

Notation ">>>" := (fn_pro).

Lemma fn_pro_nil : 
  >>> nil = nil.
Proof.
  auto.
Defined.

(* lookup is preserved under fn_pro *)

Lemma fn_pro_eq_val : forall eta n t u,
eta ^^ n = Some t ->
(>>> eta) ^^ n = Some u ->
>> t = u.
Proof.
  unfold fn_pro; induction eta, t; auto; simpl; intros; auto. 
  - inversion H; auto.
  - destruct a as [m t]; simpl in H0. 
    case_name. 
    + inversion H0; inversion H; simpl; auto. 
    + set (x0 := (<< x w)); assert (x= >> x0) as H1; auto; 
      rewrite H1; apply IHeta with n; auto. 
Defined.

Lemma fn_pro_dom_eq : forall eta,
  dom eta = dom (>>> eta).
Proof.
  induction eta as [ | nw eta0]; auto; 
  destruct nw as [n w]; simpl; rewrite IHeta0; auto.
Defined.



Lemma fn_pro_rm_comm : forall eta n,
  ((>>> eta) ^- n) = (>>> (eta ^- n)).
Proof.
  induction eta as [| a eta]; auto; simpl; intros;
  destruct a as (m, wft); destruct wft as (t, wf); auto. 
  simpl; case_name; rewrite IHeta; auto. 
Defined.

Lemma fn_pro_cons : forall n a eta,
    (>>>((n, wf_Par a) :: eta)) = ((n, Par a) :: >>> eta).
Proof.
  auto.
Defined.

Lemma fn_pro_In_dom : forall eta n u,
    >>> eta ^^ n = Some u ->
    wf_t u.
Proof.
  induction eta; simpl; intros; try discriminate.
  destruct a; destruct s; simpl in *.
  case_name. 
  - inversion H.
    rewrite <- H1; auto.
  - eapply IHeta; eauto.
Defined.

Lemma wf_t_sub_ov : forall t rho eta,
    sub_name (ov_t t) (dom eta) -> 
    wf_t (t [[>>> rho, >>> eta]]).
Proof.
  induction t; auto; simpl; intros. 
  - apply In_incl in H. generalize H; intro H1. 
    rewrite fn_pro_dom_eq in H. 
    apply lookup_exists in H. inversion H. rewrite H0.
    apply lookup_exists in H1. 
    inversion H1.  destruct x0. 
    rewrite <- fn_pro_eq_val with (eta:=eta) (n:=n) (t := (<< x0 w)); auto.
  - case (n <-? (dom (>>> rho))); intros.
    pose proof i. 
    apply lookup_exists in i. 
    inversion i. 
    rewrite <- fn_pro_dom_eq in H0. 
    apply lookup_exists in H0. inversion H0.
    destruct x0. rewrite H1.
    rewrite <- fn_pro_eq_val with (eta := rho) (n:=n) (t:=(<< x0 w)); auto. 
    apply notIn_dom_None in n0; rewrite n0; apply wPar.
  - apply wCst.
  - apply wApp; [apply IHt1 | apply IHt2]; auto with datatypes.
Defined.


Lemma open_eq_subst_t : forall t n u,
    wf_t t ->
    (t [[nil, (n, u) :: nil]]) = t.
Proof.
  induction t; simpl; intros; auto; try inversion H; try f_equal; auto.
Defined.

Hint Resolve open_eq_subst_t.
Hint Rewrite open_eq_subst_t.


Lemma subst_t_rm : 
  forall t n eta,
    (t [[nil, eta^-n]]) = (t [[nil, (n,Var n) :: eta]]).
Proof.
  induction t; auto; simpl; intros.
  - case_name.
    + rewrite env_rm_self; auto. 
    + rewrite <- env_rm_neq with (b:= n0) (rho := eta); auto. 
   - congruence. 
Qed.

Lemma subst_rm :
  forall (A : formula) (n : name) (eta : assoc trm),
    (A [nil, eta ^- n]) = (A [nil, (n, Var n) :: eta]).
Proof.
  induction A as [ [p t] | |]; auto; simpl; intros.
  - rewrite subst_t_rm; auto.
  - rewrite <- IHA1, IHA2; auto.
  - case_name.
    + rewrite <- rm_rm; auto.
    + rewrite <- IHA, rm_neq_rm; auto. 
Qed.


Lemma fn_pro_wf_t_sig : 
  forall u,
    wf_t u -> 
    (exists c, (u = Cst c) \/ (u = Par c))
    \/ (exists f u1 u2, ((u = App f u1 u2) /\ (wf_t u1) /\ (wf_t u2))).
Proof. 
  induction 1; auto; intros; [ left | left | right]; try exists x; auto. 
  destruct IHwf_t1, IHwf_t2; eauto; do 2 destruct H1, H2; 
  try do 2 destruct H1, H2; try destruct H3; try exists f, t1, t2; subst; auto.
Defined.

Lemma open_wf_trm_subst_t : 
  forall u (eta: assoc wf_trm) n t, 
    open_t (u [[nil , >>> eta ^- n]]) n t 
    = (u [[ nil , (n, t) :: >>> eta]]).
Proof.
  unfold open_t; induction u as [ m | m | m | ]; simpl; intros; auto.
  - case_name.
    + rewrite env_rm_self; simpl; case_name; auto; intuition. 
    + rewrite <- env_rm_neq; auto.
      pose proof (fn_pro_In_dom eta m) as H;
      pose proof fn_pro_wf_t_sig as H0. 
      destruct (>>> eta ^^ m). 
      * specialize H with t0; specialize H0 with t0; destruct H0; 
         try apply H; auto.
      * simpl; case_name; auto; try contradict H; intuition.
  -  f_equal; auto.
Defined.

Lemma open_wf_trm_subst :
  forall A eta n t, 
    open (A [nil,  >>> eta ^- n]) n t = ( A [nil , (n, t) :: >>> eta]).
Proof.
  unfold open; induction A as [ [P u]| | m ]; simpl; intros.
  - do 2 f_equal.  
    apply open_wf_trm_subst_t; auto. 
  - f_equal; auto.
  - f_equal; case_name. 
    * rewrite <- nil_subst; rewrite <- rm_rm; auto. 
    * rewrite rm_neq_rm; auto; rewrite fn_pro_rm_comm; auto. 
Defined.


Lemma ov_t_ind_rho :
  forall t (rho eta : assoc wf_trm),
    ov_t (t [[ >>> rho, >>> eta]]) = ov_t (t [[>>> nil, >>> eta]]).
Proof.
  induction t, rho; auto; intros.
  - destruct p as [m w]; destruct w as [t wt]; simpl. 
    case_name;  auto.  pose proof (fn_pro_In_dom rho n);
    case_env; auto. 
  - destruct p as [n w].   
    case w; intros; unfold subst; simpl;
    rewrite IHt1 with (rho := ((n, << x w0) :: rho)); 
    rewrite IHt2 with (rho := ((n, << x w0) :: rho)). 
    rewrite fn_pro_nil; auto. 
Defined. 

Lemma ov_ind_rho :
  forall A (rho eta : assoc wf_trm),
    ov (A [ >>> rho, >>> eta]) = ov (A [>>> nil, >>> eta]).
Proof.
  induction A as [ [p  t] | | ]; induction rho as [ | [ m w ] rho ]; auto; intros
  ; simpl; 
    replace ((m, >> w) :: >>> rho) with (>>> ((m , w ) :: rho)); auto. 
  - rewrite ov_t_ind_rho; auto using fn_pro_nil. 
  - rewrite IHA1, IHA2; auto using fn_pro_nil.
  - rewrite fn_pro_rm_comm; 
    rewrite IHA with (eta:= eta ^-n); auto using fn_pro_nil.
Defined. 

Lemma open_eq_subst_wf_t : forall t u eta rho n,
    wf_t u ->
    ((t [[>>>rho, >>>eta ^- n]]) [[>>> nil, (n, u) :: nil]]) 
    = (t [[>>>rho, (n, u) ::(>>> eta)]]).
Proof.
  induction t; auto; intros. 
  - rewrite <- fn_pro_nil; simpl; case_name. 
    + rewrite env_rm_self; simpl; case_name; intuition. 
    + rewrite <- env_rm_neq; auto. 
       case (n <-? (dom (>>> eta))) as [H1 | H1]. 
      * generalize H1; intro H2. rewrite <- fn_pro_dom_eq in H2.
        apply lookup_exists in H1; destruct H1 as [t St]; 
        apply lookup_exists in H2; destruct H2 as [tt Stt];
        rewrite St.
        assert (H0 := fn_pro_eq_val eta n Stt St). 
        destruct tt as [tt wt]; simpl in H0; subst. 
        case wt; intros; auto; simpl.  
        rewrite! open_eq_subst_t; auto. 
      * pose proof (notIn_dom_None n (>>> eta) H1) as H0;
        rewrite H0; simpl. case_name; intuition.
  - rewrite! open_eq_subst_t; auto. 
    rewrite fn_pro_rm_comm; auto; simpl; case_Env.
    + apply fn_pro_In_dom with rho n; auto.
    + constructor.
  -  simpl; rewrite IHt1, IHt2; auto. 
Defined.
 
Lemma open_eq_subst : forall A eta rho t n,
    wf_t t ->
    ((A [>>>rho, >>>eta ^- n]) [>>> nil, (n, t) :: nil]) 
    = (A [>>>rho, (n, t) ::(>>> eta)]).
Proof. 
  induction A as [ [p t] | |]; auto; intros; simpl. 
  - do 2 f_equal; apply open_eq_subst_wf_t; auto.
  - rewrite IHA1, IHA2; auto. 
  - f_equal. case_name; auto.  
    + rewrite <- rm_rm, nil_subst; auto. 
    + rewrite <- rm_neq_rm; auto; rewrite fn_pro_rm_comm;
      rewrite IHA with (eta := eta ^- n); auto.
Defined. 

Lemma wf_trm_subst_wf_t : 
  forall t (rho eta : assoc wf_trm),
    sub_name (ov_t t) (dom eta)
    -> ov_t (t[[ >>> rho, >>> eta]])= nil.
Proof.
  induction t; intros. 
  - simpl in H; pose proof (fn_pro_In_dom eta n); apply In_incl in H; 
    rewrite fn_pro_dom_eq in H.
    pose proof (lookup_exists n (>>> eta) H).
    simpl; destruct H1 as [t St]; rewrite St. 
    apply wf_t_ov_empty; apply H0; auto. 
  - rewrite ov_t_ind_rho, fn_pro_nil; auto. 
  - rewrite ov_t_ind_rho, fn_pro_nil; auto. 
  - simpl; rewrite IHt1, IHt2; simpl in H; auto with datatypes.   
Defined.    


Lemma ov_wf_trm_subst : forall A (eta : assoc wf_trm), 
    sub_name (ov A) (dom eta) ->
    wf (A [nil, >>> eta]).
Proof.      
  induction A as [ [p t] | |]; simpl; intros eta Hsub.
  - apply wAtom;  auto with datatypes. 
    rewrite <- fn_pro_nil. apply wf_t_sub_ov; auto. 
  - apply wImply; auto with datatypes.
  - set (a := new (op (A [nil,>>> eta ^- n]))).
    apply wForall with (c := a);
      [ apply new_fresh
      | rewrite (open_wf_trm_subst); rewrite <- fn_pro_cons; apply IHA; 
        simpl; apply incl_rm_2; auto]. 
Defined.

Lemma wf_trm_subst_wf: 
  forall A (rho eta : assoc wf_trm),
     sub_name (ov A) (dom eta)
    -> ov ( A [ >>> rho, >>> eta])= nil.
Proof.
  intros; rewrite ov_ind_rho; apply wf_ov_empty, ov_wf_trm_subst; auto. 
Defined.
